/**
 * BloodEffect.cpp: implementation of the BloodEffects
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "BloodEffect.h"
#include "Engine.h"
#include "Renderer.h"
#include "Background.h"
#include "Clock.h"
#include "Vector2D.h"
#include <cstdlib>
#include <cmath>
#include <GL/gl.h>
//#include <iostream>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH

const int N_PARTICLES = 500;
const float START_Z = 1.0f;
const float BASE_VELOCITY = 300.f / Clock::LOGICFRAMES_PER_SECOND;
const float GRAVITY = 9;
const float DEG_TO_RAD = PI/180;
const float BLOOD_INTENSITY = 0.66f; // intensity of a bloodpixel on the ground

BloodEffect::BloodEffect(const Vector2D& pos, const Vector2D& dir)
	: particles(N_PARTICLES)
{
	float angleRad = dir.toDegrees() * DEG_TO_RAD;
	float dAngleRad = 0;
	float dVelocity = 0;
	foreach (Particle& p, particles)
	{
		switch (rand() % 6)
		{
			case 5: dAngleRad = (60 - (rand() % 120)); break;
			case 4: dAngleRad = (40 - (rand() % 80)); break;
			case 3:
			case 2:
			case 1:
			case 0: dAngleRad = (20 - (rand() % 40)); break;
		}
		dAngleRad *=  DEG_TO_RAD;
		dVelocity = (150 - (rand() % 300)) / Clock::LOGICFRAMES_PER_SECOND;

		p.pos.x = pos.x;
		p.pos.y = pos.y;
		p.pos.z = START_Z;

		p.vel.x = cos(angleRad + dAngleRad) * (BASE_VELOCITY + dVelocity);
		p.vel.y = sin(angleRad + dAngleRad) * (BASE_VELOCITY + dVelocity);
		p.vel.z = BASE_VELOCITY + dVelocity;
	}
}

BloodEffect::~BloodEffect()
{
//	std::cout << "BloodEffect " << this << " dying" << std::endl;
}


void BloodEffect::update()
{
	for (ParticleIterator p = particles.begin(); p != particles.end(); ++p)
	{
		p->pos += p->vel;
		p->vel.z -= GRAVITY;

		if (p != particles.end() && p->pos.z + p->vel.z <= 0.0)
		{
			// we're getting more random patterns if we use the coordinates
			// of ANOTHER particle instead of the ones of the dying particle
			ParticleIterator x = p;
			++x;
			if (x == particles.end())
				x = p;

			Renderer::get().getBackground().addBlood(x->pos.getX(), x->pos.getY(), BLOOD_INTENSITY);
			p = particles.erase(p);

			if (p == particles.end())
				break;
		}
	}
}

void BloodEffect::draw(float dt)
{
//	float zmin = 1000, zmax = -1000;
	float intensity = 0;
	glBegin(GL_POINTS);
	foreach (const BloodEffect::Particle& p, particles)
	{
		intensity = 0.5;

		// we filter over a 3x3 area :)
		glColor4f(1.0, 0, 0, (intensity / FILTERFACTOR));
		glVertex3f(p.pos.x - 1, p.pos.y - 1, p.pos.z);
		glVertex3f(p.pos.x, p.pos.y - 1, p.pos.z);
		glVertex3f(p.pos.x + 1, p.pos.y - 1, p.pos.z);

		glVertex3f(p.pos.x - 1, p.pos.y, p.pos.z);
		glColor4f(1.0, 0, 0, intensity);
		glVertex3f(p.pos.x, p.pos.y, p.pos.z);
		glColor4f(1.0, 0, 0, (intensity / FILTERFACTOR));
		glVertex3f(p.pos.x + 1, p.pos.y, p.pos.z);

		glVertex3f(p.pos.x - 1, p.pos.y + 1, p.pos.z);
		glVertex3f(p.pos.x, p.pos.y + 1, p.pos.z);
		glVertex3f(p.pos.x + 1, p.pos.y + 1, p.pos.z);

//		glVertex3f(p.pos.x + p.vel.x, p.pos.y + p.vel.y, p.pos.z + p.vel.z);
//		if (p.pos.z > zmax) zmax = p.pos.z;
//		if (p.pos.z < zmin) zmin = p.pos.z;
	}
	glEnd();
	glColor4f(0, 0, 0, 1);
//	std::cout << zmin << '/' << zmax << '\n';
}


bool BloodEffect::isFinished()
{
	return particles.empty();
}



