/**
 * Engine.cpp:
 * 	a SDL-specific implementation of Engine.h
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "Engine.h"
#include "Effect.h"
#include "Game.h"
#include "Renderer.h"
#include "Clock.h"
#include "EngineError.h"
#include "SpriteManager.h"
#include "Cursor.h"
#include "SoundManager.h"
#include "Vector2D.h"
#include "PlayerPawn.h"
#include "ResourceManager.h"
#include <SDL/SDL.h>
#include <cstdio>
#include <ctime>
//#include <iostream>


Engine::Engine()
	: isRunning(false),
	  windowWidth(WORLDWIDTH),
	  windowHeight(WORLDHEIGHT)
{
}

Engine::~Engine()
{
//	std::cout << "Engine dying\n";
	tearDown();
}

Engine& Engine::get()
{
	static Engine e;
	return e;
}


void Engine::setup()
{
	if (SDL_Init(0) < 0)
		throw EngineError("unable to initialize SDL");

	atexit(SDL_Quit);
	resourceManager.reset(new ResourceManager());
	ResourceManager::instance = resourceManager.get();
	Renderer::instance = new Renderer();
	Renderer::get().setup(windowWidth, windowHeight);
	spriteManager.reset(new SpriteManager());
	SoundManager::instance = new SoundManager();
	SDL_WM_SetCaption("ZombieShooter", 0);
	srand(time(0));
	game.reset(new Game());
	Game::instance = game.get();
}

void Engine::run()
{
	setup();
	isRunning = true;
	Clock& clock = Clock::get();
	Cursor cursor;
	cursor.setShootingCursor();
	game->start();
	while (isRunning)
	{
		while (clock.needLogicTick())
		{
			clock.newLogicTick();
			if (game->isRunning())
				game->update();
			handleInput();
			handleEffects();
//			std::cout << clock.getFps() << '\n';
		}
		clock.newDrawTick();
		Renderer::get().draw(clock.getProgress());
		SDL_Delay(10); // we don't need all the cpu time :)
	}
}


/*
 * Does some cleaning up.
 * NOTE: the order in which this is done is important, so that there are
 * no circular references left. I.e. we can't kill the renderer as long
 * as there are still objects that need to unregister themselves from it
 */
void Engine::tearDown()
{
	spriteManager.reset();
	effects.clear();
	delete Renderer::instance;
	delete SoundManager::instance;

	SDL_Quit();
//	std::cout << "engine: " << objects.size() << '\n';
//	std::cout << "renderer: " << Renderer::get().animators.size() << '\n';
}


void Engine::handleInput()
{
	PlayerPawn& player = game->getPlayerPawn();
	handleKeypress();
	int mouseX, mouseY;
	SDL_GetMouseState(&mouseX, &mouseY);
	mouseY = windowHeight-mouseY;
	Vector2D pos = player.getPosition() + player.getSize() / 2;
	Vector2D mouse(static_cast<float>(mouseX), static_cast<float>(mouseY));
	player.setLookingDirection(mouse - pos);

	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
		switch (event.type)
		{
			case SDL_QUIT:
				isRunning = false;
				break;

			case SDL_VIDEORESIZE:
				windowWidth = event.resize.w;
				windowHeight = event.resize.h;
				Renderer::get().resize(windowWidth, windowHeight);
				break;

			case SDL_MOUSEBUTTONDOWN:
				if (event.button.button == SDL_BUTTON_LEFT)
					player.shoot();

			case SDL_KEYDOWN:
				if (event.key.keysym.sym == SDLK_ESCAPE)
					isRunning = false;
				break;
		}
	}
}


void Engine::handleKeypress()
{
	Uint8 *keystate = SDL_GetKeyState(NULL);
	char dirX = 0;
	char dirY = 0;
	if (keystate[SDLK_UP] || keystate[SDLK_w])
		dirY = 1;
	else if (keystate[SDLK_DOWN] || keystate[SDLK_s])
		dirY = -1;
	if (keystate[SDLK_RIGHT] || keystate[SDLK_d])
		dirX = 1;
	else if (keystate[SDLK_LEFT] || keystate[SDLK_a])
		dirX = -1;
	Vector2D dir(dirX, dirY);
	dir.normalize();
	game->getPlayerPawn().setDirection(dir);
}

void Engine::registerEffect(Effect* object)
{
	effects.push_back(EffectPtr(object));
}



void Engine::handleEffects()
{
	for (EffectList::iterator i = effects.begin(); i != effects.end(); ++i)
	{
		(*i)->update();
		if ((*i)->isFinished())
			i = effects.erase(i);
	}
}
