#include "ResourceManager.h"
#include "EngineError.h"
#include <SDL/SDL.h>
#include <GL/gl.h>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH

//#include <iostream>


ResourceManager* ResourceManager::instance = 0;

ResourceManager::ResourceManager()
{
}

void ResourceManager::loadResources()
{
	samples[PISTOL_SHOT] = loadSample("data/shot.ogg");
	samples[PISTOL_HIT] = loadSample("data/hit.ogg");
//	samples[ZOMBIE_DEAD]= loadSample("data/zombie_dead.wav");
//	samples[ZOMBIE_DEAD] = loadSample("data/silence.wav");
	samples[PLAYER_SWEAR] = loadSample("data/player_swear.ogg");
	samples[PLAYER_DEAD]= loadSample("data/player_dead.ogg");
	samples[LEVEL_UP] = loadSample("data/level_up.ogg");

	const char* textureArrayPaths[] = {"data/player.bmp",
		"data/zombie.bmp", "data/shot.bmp", "data/digits.bmp"};

	foreach (const char* path, textureArrayPaths)
		textureArrays[path] = loadTextures(path);

	const char* texturePaths[] = {"data/ground.bmp", "data/game_over.bmp"};
	foreach (const char* path, texturePaths)
	{
		SDL_Surface* tmp = loadSurface(path);
		textures[path] = convertToOpenGlTexture(tmp);
	}
}


ResourceManager::~ResourceManager()
{
	for (SampleMap::iterator i = samples.begin(); i != samples.end(); ++i)
	{
		SDL_FreeRW(i->second);
	}
}


SDL_RWops* ResourceManager::getSample(Sample s) const
{
	SampleMap::const_iterator tmp = samples.find(s);
	if (tmp != samples.end())
		return tmp->second;
	else
		throw EngineError(std::string("Unknown sample: ") + (char) s);
}


Texture ResourceManager::getTexture(const char* path) const
{
	TextureMap::const_iterator tmp = textures.find(path);
	if (tmp != textures.end())
		return tmp->second;
	else
		throw EngineError(std::string("Unknown texture: ") + path);
}


TextureArray ResourceManager::getTextures(const char* path) const
{
	TextureArrayMap::const_iterator tmp = textureArrays.find(path);
	if (tmp != textureArrays.end())
		return tmp->second;
	else
		throw EngineError(std::string("Unknown TextureArray: ") + path);
}



SDL_RWops* ResourceManager::loadSample(const char* path)
{
	SDL_RWops* tmp = SDL_RWFromFile(path, "rb");
	if (!tmp)
	{
		throw EngineError(std::string("unable to load sound from ") + path);
	}

	return tmp;
}




SDL_Surface* ResourceManager::loadSurface(const char* path)
{
	SDL_Surface* surface = SDL_LoadBMP(path);
	if (!surface)
		throw EngineError(std::string("unable to load texture from ") + path);

	return surface;
}



Texture ResourceManager::convertToOpenGlTexture(SDL_Surface* surface)
{
	SDL_Rect rect = {0, 0, surface->w, surface->h};
	return convertToOpenGlTexture(surface, rect);
}


Texture ResourceManager::convertToOpenGlTexture(SDL_Surface* surface, SDL_Rect& srcRect)
{
	// NOTE: for some reason the comiler won't accept GL_BGR(A)
	// so right now the textures are wrongly colored.
	// NOTE ALSO: the SDL-surface we create later on will allways have
	// mode GL_RGBA and 24 bpp.
	int mode;
	if (surface->format->BytesPerPixel == 3)
		mode = GL_RGB;
	else if (surface->format->BytesPerPixel == 4)
		mode = GL_RGBA;
	else
	{
		throw EngineError("Textures need to have 24 or 32 bpp");
	}

	// OK, we always use RGBA, so we have 4 Bytes per pixel and use
	// RGBA
	int bpp = 4*8; //surface->format->BitsPerPixel,
	mode = GL_RGBA;

	// (255, 0, 255) is used as the transparent color.
	SDL_SetColorKey(surface, SDL_SRCCOLORKEY, SDL_MapRGB(surface->format, 255, 0, 255));

	// I have no idea why i do this, but transparency doesn't work if I don't ?-)
	if ( (surface->flags & SDL_SRCALPHA) == SDL_SRCALPHA )
		SDL_SetAlpha(surface, 0, SDL_ALPHA_TRANSPARENT);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	SDL_Surface* spriteSurface = SDL_CreateRGBSurface(
		SDL_SWSURFACE | SDL_SRCCOLORKEY, srcRect.w, srcRect.h, bpp,
		// make sure to use RGBA, as (at least under Windows) I can't get GL_ABGR to work :(
		#if SDL_BYTEORDER == SDL_LIL_ENDIAN
			       0x000000FF,
			       0x0000FF00,
			       0x00FF0000,
			       0xFF000000
		#else
			       0xFF000000,
			       0x00FF0000,
			       0x0000FF00,
			       0x000000FF
		#endif
       );

	SDL_FillRect(spriteSurface, 0, 0);
	SDL_BlitSurface(surface, &srcRect, spriteSurface, 0);

	Texture textureid;
	glGenTextures(1, &textureid);
	glBindTexture(GL_TEXTURE_2D, textureid);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
		spriteSurface->w, spriteSurface->h, 0, mode,
		GL_UNSIGNED_BYTE, spriteSurface->pixels);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	SDL_FreeSurface(spriteSurface);

//	printf("%d\n", textureid);
	return textureid;
}



TextureArray ResourceManager::loadTextures(const char* file)
{
	SDL_Surface* s = loadSurface(file);

	// assume the image contains several sprites of HEIGHTxHEIGHT
	// copy each such sprite in a new surface and create a texture out of it
	int size = s->h;
	if ((size & (size - 1)) != 0)
	{
		throw EngineError("Texture sizes need to be powers of 2: ", file);
	}

	SDL_Rect srcRect = {0, 0, size, size};
	int n = s->w / size;
	TextureArray retval;
	for (int i = 0; i < n; ++i)
	{
		srcRect.x = i * size;
		try
		{
			retval.push_back(convertToOpenGlTexture(s, srcRect));
		}
		catch (EngineError& e)
		{
			SDL_FreeSurface(s);
			throw e;
		}
	}
	SDL_FreeSurface(s);
	return retval;
}


